#
# Robert Olson olson@mcs.anl.gov
# Argonne National Laboratory (c) 1996
#

puts "loading web code"

option add Vic.webInterval 1000

proc init_web {} {
    global Web
#    puts "Create listener"
    set sock [socket -server new_connection  2500 ]
#    puts "Got listener $sock"
    puts "Configuration: [fconfigure $sock]"
    if { $sock == "" } {
	puts "Falling back to host-selected port"
	set sock [socket -server new_connection 0 ]
    }

    if { $sock == "" } {
	puts "Cannot initialize web"
	break;
    }
    
    set err [ catch { set Web(port) [fconfigure $sock -sockname ] } ]
    if { $err != 0 } {
	puts "configure fails"
	set Web(port) 2500
    }

    set Web(listener) $sock
#    puts "listener: $Web(listener) port $Web(port)"
    
}

proc new_connection { name addr port } {
    global client_hits
    
#    puts "new connection $name $addr $port"

    if [ info exists client_hits($addr) ] { 
	incr client_hits($addr)
    } else {
	set client_hits($addr) 1
    }

    fconfigure $name -translation binary
    
    handle_request $name
}

proc handle_request { conn } {
    set path ""
    while { 1 } {
	set line [gets $conn]
	set line [string trimright $line "\r"]
	if { [string compare $line ""] == 0 } {
	    break
	}
	set comps [split $line]

	set cmd [lindex $comps 0]

#	puts "Read '$line' '$cmd'"

	case $cmd {
	    "GET" { set path [lindex $comps 1] }
	    "*:" {
		set hdr [string trimright $cmd ":"]
		set args [lrange $comps 1 end]
		set headers($hdr) $args
#		puts "got hdr <$hdr> <$args>"
	    }
	    default { puts "got other" }
	}
    }
    
    handle_get $conn $path headers
}

#
# Handle the get request.
#
# We handle the following:
#
#    /		Return a page with links to the single-shot and
#		pushed video streams available
#
#    /stream/ident	Start up a server-push stream for source 'ident'.
#    /image/ident	Return the next available image for 'ident'.
#
proc handle_get { conn path headers_nm } {

    upvar $headers_nm headers

    foreach hdr [array names headers] {
	set val $headers($hdr)
#	puts "<$hdr> = <$val>"
    }

    case $path {
	"/" {
	    handle_index $conn headers
	}
	"/stream/*" {
	    set ident [string range $path 8 end]
	    handle_video_request $conn stream $ident headers
	}
	"/image/*" {
	    set ident [string range $path 7 end]
	    handle_video_request $conn image $ident headers
	}
	default {
	    handle_invalid_path $conn $path headers
	}
    }
}

#
# Handle a request.
#
# Register this connection with the output device, marking the
# registration as one-shot or persistent.
# 
proc handle_video_request { conn type ident headers_nm } {

    set one_shot 0
    if { $type == "image" } {
	set one_shot 1
    }

    if { [regexp {^(_o[0-9]*)/.*\.(.*)$} $ident match obj img_type] } {

	if { $one_shot } {
	    handle_oneshot $conn $obj $img_type
	} else {
	    handle_stream $conn $obj $img_type
	}
    } else {
	puts "match failed"
	handle_invalid_path $conn $ident headers_nm
    }
}

proc handle_oneshot { conn obj img_type } {
    
    puts $conn "HTTP/1.0 200 Document Follows"
    puts $conn "Server: tclstuff"
    puts $conn "Content-type: image/gif"

    $obj request_oneshot $img_type $conn
}

proc handle_stream { conn obj img_type } {

    set separator "imagesep[clock clicks]"

    puts $conn "HTTP/1.0 200 Document Follows"
    puts $conn "Server: tclstuff"
    puts $conn "Content-type: multipart/x-mixed-replace;boundary=$separator"
    puts $conn ""
    puts $conn "--$separator"

    fconfigure $conn -blocking 0
    fconfigure $conn -buffering none
    fileevent $conn writable [list stream_writable $obj $conn]
    $obj request_stream $img_type $conn $separator
}

proc stream_writable { obj conn } {
#    puts "stream is writable $obj $conn"
    $obj writable $conn
}

proc handle_index { conn headers_nm } {
    global client_hits
    global outputDeviceList V

    send_html_header $conn
    send_html_head $conn "Vic video index"
    send_html_body_start $conn

    #
    # Vic session stuff
    #

    set net $V(data-net)
    
    puts $conn "<h1>Vic on [info hostname]</h1>"

    puts $conn "Destination: [$net addr]/[$net port]<br>"
    puts $conn "Vic owned by [[srctab local] sdes name]<p>"

    puts $conn "Current sources:<p>"
    puts $conn "<ul>"
    srctab gen-init
    while { [set src [srctab gen-next]] != "" } {
	puts "got src $src"
	set stats [get_stats $src]
	set fps [lindex $stats 0]
	set bps [lindex $stats 1]
	set loss [lindex $stats 2]
	set who [$src sdes name]
	set addr [ $src addr]
	puts $conn "<li> $who @ $addr <br>"
	puts $conn "$fps $bps $loss"
    }
    puts $conn "</ul>"

    puts $conn "<h3>Available streams:</h3>"

    puts $conn "<ul>"
    foreach source $outputDeviceList {
	set src_name [$source nickname]
	puts "Have $source $src_name"
	set img_url "/image/$source/video.gif"
	set img_link [make_link $img_url "Single image from $src_name"]
	set stream_url "/stream/$source/video.gif"
	set stream_link [make_link $stream_url "Stream from $src_name"]
	puts $conn "<li> $img_link"
	puts $conn "<li> $stream_link"
    }
    puts $conn "</ul>"

    puts $conn "Who's been using me?"
    puts $conn "<ul>"
    foreach host [array names client_hits] {
	puts $conn "<li> $host: $client_hits($host) requests"
    }
    puts $conn "</ul>"

    send_html_body_end $conn

    close $conn
}

proc get_stats { src } {
	global ftext btext ltext fpshat bpshat lhat shat V

	set key $src
	if { "$src" != "$V(session)" } {
		# only compute loss statistic for network side
		set p [$src np]
		set s [$src ns]
		set shat($key) $s
		set lhat($key) [expr $s-$p]
		if {$shat($key) <= 0.} {
			set loss 0
		} else {
			set loss [expr 100*$lhat($key)/$shat($key)]
		}
		if {$loss < .1} {
		    set loss (0%)
		} elseif {$loss < 9.9} {
		    set loss  [format "(%.1f%%)" $loss]
		} else {
		    set loss [format "(%.0f%%)" $loss]
		}
	}
	set bpshat($key) [$src nb]
	set fpshat($key) [$src nf]

	set fps $fpshat($key)
	set bps $bpshat($key)
	
	if { $fps < .1 } {
		set fps "0 f/s"
	} elseif { $fps < 10 } {
		set fps [format "%.1f f/s" $fps]
	} else {
		set fps [format "%2.0f f/s" $fps]
	}
	if { $bps < 1 } {
		set bps "0 bps"
	} elseif { $bps < 1000 } {
		set bps [format "%3.0f bps" $bps]
	} elseif { $bps < 1000000 } {
		set bps [format "%3.0f kb/s" [expr $bps / 1000]]
	} else {
		set bps [format "%.1f Mb/s" [expr $bps / 1000000]]
	}
	return [list $fps $bps $loss]
}

proc handle_invalid_path { conn path headers_nm } {
    send_html_header $conn
    send_html_head $conn "Not found"
    send_html_body_start $conn
    puts $conn "<h1>Not found</h1>"
    puts $conn "Path <tt>$path</tt> not found."
    send_html_body_end $conn
    close $conn
}

proc make_link { url text } {
    return "<a href=\"$url\">$text</a>"
}

proc send_html_head { conn title } {
    puts $conn "<html> <head>"
    puts $conn "<title>$title</title>"
    puts $conn "</head>"
}

proc send_html_body_start { conn } {
    puts $conn "<body>"
}

proc send_html_body_end { conn } {
    puts $conn "</body>"
    puts $conn "</html>"
}

proc send_html_header { conn } {
    puts $conn "HTTP/1.0 200 Document Follows"
    puts $conn "Server: tclstuff"
    puts $conn "Content-type: text/html"
    puts $conn ""
}

proc send_plain_header { conn } {
    puts $conn "HTTP/1.0 200 Document Follows"
    puts $conn "Server: tclstuff"
    puts $conn "Content-type: text/plain"
    puts $conn ""
}

#proc bgerror { msg } {
#n    puts "background error: $msg"
#}

#init_web
#
#while {1} {
#    update
#}

